#ifndef ENGINE_H
#define ENGINE_H

#include <containers/vec_.h>
#include <containers/hash_.h>
#include <containers/tree.h>
#include <math/matrix.h>
#include <math/vector.h>
#include <sys/surface.h>
#include <sys/nativedl.h>
#include <sys/dlist.h>
#include <dltrans.h>
#include <viewport.h>
#include <node/xformer.h>
#include <node/mesh.h>
#include <node/billboard.h>
#include <node/lux.h>
#include <node/camera.h>

typedef enum {
	OBJECT_RAW,
	OBJECT_SCENE,
	OBJECT_MESH,
	OBJECT_CAMERA,
	OBJECT_BILLBOARD,
	OBJECT_LUX
} ObjectType;

typedef struct {
	ObjectType	tag;
	Xformer*	xf;
} Raw_Xformer;

typedef struct {
	ObjectType	tag;
	Xformer*	xf;
	Mesh*		mesh;
} Mesh_Xformer;

typedef struct {
	ObjectType	tag;
	Xformer*	xf;
	Billboard*	billboard;
} Billboard_Xformer;

typedef struct {
	ObjectType	tag;
	Xformer*	xf;
	Camera*		camera;
} Camera_Xformer;

typedef struct {
	ObjectType	tag;
	Xformer*	xf;
	Lux*		lux;
} Lux_Xformer;

typedef struct {
	Vec			lights;
	const Viewport*		viewport;
	const Camera_Xformer*	camera_xf;
	const Surface*		renderTarget;
	const Surface*		zBuffer;
	Vec			cpuMatrixes;
	Matrix			projMtxF;
	DList			dl;
	DLTrans			dltrans;
	NativeDL		rsp_cfb;
	NativeDL		rsp_zb;
	Hash_			mesh_to_vpairs;

	Vec			xformers;
	Vec			meshes;
	Vec			billboards;
	Vec			cameras;
	Vec			luxes;
} Engine;

#include "task.h"

void init_Engine( Engine* engine );
void setViewport_Engine( Engine* engine, const Viewport* viewport );
void setCamera_Engine( Engine* engine, const Camera_Xformer* camera_xf );
void setRenderTarget_Engine( Engine* engine, const Surface* renderTarget );
void setZBuffer_Engine( Engine* engine, const Surface* zbuffer );
const Viewport* getViewport_Engine( const Engine* engine );
const Camera_Xformer* getCamera_Engine( const Engine* engine );
const Surface* getRenderTarget_Engine( const Engine* engine );
const Surface* getZBuffer_Engine( const Engine* engine );
void addLight_Engine( Engine* engine, const Lux* light );
void removeLight_Engine( Engine* engine, const Lux* light );
void processTask_Engine( Engine* engine, const Task* task );

void sceneTaskF_Engine( Engine* engine, TreeNode* scene );
void overlayTaskF_Engine( Engine* engine, void* arg );
void clearColorbufF_Engine( Engine* engine, void* arg );
void clearZbufF_Engine( Engine* engine, void* arg );

Xformer* createXformer_Engine( Engine* self, ProcessFunc_Xformer processor );
Raw_Xformer createRaw_Engine( Engine* self );
Mesh_Xformer createMesh_Engine( Engine* self, StaticMesh const* static_mesh );
Billboard_Xformer createBillboard_Engine( Engine* self, StaticMesh const* static_mesh );
Camera_Xformer createCamera_Engine( Engine* self, CameraType type );
Lux_Xformer createLux_Engine( Engine* self );

#endif	/* ENGINE_H */

