#include <abort.h>
#include <float.h>
#include <sys/region.h>
#include <node/xformer_funcs.h>
#include "engine.h"
#include "rectangle.h"

#define MAX_NODES	64

static bool mesh_is_opaque( const Mesh* mesh ) {
	return 1.0f <= mesh->opacity;
}

static bool mesh_to_vpairs_equ( void* key1, void* key2 ) {
	return key1 == key2;
}

void init_Engine( Engine* engine ) {
	init_Vec( &engine->lights, sizeof(Lux*), ALIGN_DEFAULT );
	init_Vec( &engine->cpuMatrixes, sizeof(Matrix), ALIGN_4B );
	resize_Vec( &engine->cpuMatrixes, MAX_NODES );
	hash_Construct(
		&engine->mesh_to_vpairs,
		sizeof(Mesh*),
		sizeof(Vec*),
		mesh_to_vpairs_equ
	);
	init_DList( &engine->dl );

	init_NativeDL( &engine->rsp_cfb );
	init_NativeDL( &engine->rsp_zb );

	//
	init_Vec(
		&engine->xformers,
		sizeof(Xformer),
		ALIGN_DEFAULT			// how to get a structure's alignment?
	);
	resize_Vec( &engine->xformers, 30 );	// don't mess with the vec from now - we don't want any reallocations to occur

	init_Vec(
		&engine->meshes,
		sizeof(Mesh),
		ALIGN_DEFAULT
	);
	resize_Vec( &engine->meshes, 30 );

	init_Vec(
		&engine->billboards,
		sizeof(Billboard),
		ALIGN_DEFAULT
	);
	resize_Vec( &engine->billboards, 30 );

	init_Vec(
		&engine->cameras,
		sizeof(Camera),
		ALIGN_DEFAULT
	);
	resize_Vec( &engine->cameras, 30 );

	init_Vec(
		&engine->luxes,
		sizeof(Lux),
		ALIGN_DEFAULT
	);
	resize_Vec( &engine->luxes, 30 );
}

inline void setViewport_Engine( Engine* engine, const Viewport* viewport ) {
	engine->viewport = viewport;
}

inline void setCamera_Engine( Engine* engine, const Camera_Xformer* camera_xf ) {
	engine->camera_xf = camera_xf;
}

inline void setRenderTarget_Engine( Engine* engine, const Surface* renderTarget ) {
	engine->renderTarget = renderTarget;
}

inline void setZBuffer_Engine( Engine* engine, const Surface* zbuffer ) {
	engine->zBuffer = zbuffer;
}

inline const Viewport* getViewport_Engine( const Engine* engine ) {
	return engine->viewport;
}

inline const Camera_Xformer* getCamera_Engine( const Engine* engine ) {
	return engine->camera_xf;
}

inline const Surface* getRenderTarget_Engine( const Engine* engine ) {
	return engine->renderTarget;
}

inline const Surface* getZBuffer_Engine( const Engine* engine ) {
	return engine->zBuffer;
}

inline void addLight_Engine( Engine* engine, const Lux* light ) {
	pushBack_Vec( &engine->lights, &light );
}

void removeLight_Engine( Engine* engine, const Lux* light ) {
	if(engine) {}
	if(light) {}
}

static void handleMesh( DList* dlist, Mesh_Xformer const* mesh_xf, Camera_Xformer const* camera_xf ) {
	pushLoadMatrix_DList(
		dlist,
		getResultMatrix_Xformer( mesh_xf->xf ),
		MATRIX_MODELVIEW
	);

	const Gfx* meshdl;
	const Layer* layer;
	const Rectangle* tile;
	Material mat;
	CombineMode prev_cm = -1;
	CombineModeAlpha prev_cma = -1;
	RenderMode prev_rm = -1;

	for( uint32_t i=0, n=mesh_xf->mesh->num_layers; i<n; ++i ) {
		layer = &mesh_xf->mesh->layers[i];
		meshdl = layer->gfx;
		tile = &layer->tile;

		///if( NULL == layer->mat || NULL == meshdl )
		///	abort();

		clone_Material(
			&mesh_xf->mesh->materials[ layer->material_id ],
			&mat
		);
		getColor_Material( &mat ) -> t *= getOpacity_Mesh( mesh_xf -> mesh );
		pushSetMaterial_DList( dlist, &mat );

		CombineMode cm;
		if( TEXTURE_ID_NONE != mat.tex_id ) {
			Texture* tex = &mesh_xf->mesh->textures[mat.tex_id];
			Rectangle tttile = {
				0,
				0,
				tex->width - 1,
				tex->height - 1
			};

			if( (tile->p0x == 0) && (tile->p0y == 0) && (tile->p1x == 0) && (tile->p0y == 0) ) {		// HACKY
				pushLoadTexture_DList( dlist, tex, tttile, false );
			} else {
				pushLoadTexture_DList( dlist, tex, *tile, true );
			}
			cm = CM_DECAL;

			if( TEXTUREMAPPING_ENV == mat.tmapping ) {
				pushSetTextureScale_DList( dlist, 0.125f, 0.125f );
				pushSetGeometryMode_DList( dlist, GM_LIGHTING | GM_TEXGEN );
				Coordsys const* camcs = getCoordsys_Xformer( camera_xf -> xf );
				Coordsys const* meshcs = getCoordsys_Xformer( mesh_xf -> xf );
				
				pushSetLookAt_DList( dlist, &camcs->origin, &meshcs->origin, &camcs->v );
			}
		} else {
			cm = CM_SHADE;
		}

		CombineModeAlpha cma = (getColor_Material( &mat ) -> t == 1.0f) ? CMA_OPAQUE : CMA_TRANSLUCENT;
		RenderMode rm = (getColor_Material( &mat ) -> t == 1.0f) ? RM_ZB_TEX_EDGE : RM_ZB_TRANSLUCENT;	// ? ZB_OPAQUE
		if( prev_rm != rm ) {
			pushSetRenderMode_DList( dlist, rm );
			prev_rm = rm;
		}
		if( prev_cm != cm || prev_cma != cma ) {
			pushSetCombineMode_DList( dlist, cm, cma );
			prev_cm = cm;
			prev_cma = cma;
		}

		pushSetSegment_DList( dlist, 0x0, (void*)layer->vts_ofs );
		pushJumpToTranslated_DList( dlist, meshdl );
	}
}

static void handleBillboard( DList* dlist, Billboard_Xformer const* bb_xf, Camera_Xformer const* camera_xf ) {
	handleMesh( dlist, (Mesh_Xformer*)bb_xf, camera_xf );
}

static void resetState( DList* dlist ) {
	pushClearGeometryMode_DList( dlist, GM_LIGHTING | GM_TEXGEN );
	pushSetGeometryMode_DList(
		dlist,
		GM_CULLBACK |
		GM_ZBUFFER |
		GM_LIGHTING |
		GM_SHADE |
		GM_GOURAUD
	);
	pushSetTextureScale_DList( dlist, 1.0f, 1.0f );
	pushPipeSync_DList( dlist );
	pushSetRenderMode_DList( dlist, RM_ZB_OPAQUE );
}

void sceneTaskF_Engine( Engine* engine, TreeNode* scene ) {
	const Viewport* vp = getViewport_Engine( engine );
	const Camera_Xformer* cam_xf = getCamera_Engine( engine );
	switch( cam_xf->camera->type ) {
		case PERSPECTIVE_CAMERA: {
			float perspNorm;
			perspective_Matrix(
				&engine->projMtxF,
				&perspNorm,
				getPerspFOVY_Camera( cam_xf->camera ),
				getWidth_Viewport( vp ) / getHeight_Viewport( vp ),
				getPerspNear_Camera( cam_xf->camera ),
				getPerspFar_Camera( cam_xf->camera )
			);
			pushNormalizePerspective_DList( &engine->dl,
				perspNorm
			);

			break;
		}
		case ORTHOGRAPHIC_CAMERA: {
			orthographic_Matrix(
				&engine->projMtxF,
				getOrthoLeft_Camera( cam_xf->camera ),
				getOrthoRight_Camera( cam_xf->camera ),
				getOrthoBottom_Camera( cam_xf->camera ),
				getOrthoTop_Camera( cam_xf->camera ),
				getOrthoNear_Camera( cam_xf->camera ),
				getOrthoFar_Camera( cam_xf->camera )
			);
			break;
		}
		default:
			abort();
			break;
	}

	const Matrix* mtx = getMatrix_Xformer( cam_xf->xf );
	Matrix inv;
	invert_Matrix( mtx, &inv );
	multiply_Matrix( &engine->projMtxF, &inv, &engine->projMtxF );
	pushLoadMatrix_DList(
		&engine->dl,
		&engine->projMtxF,
		MATRIX_PROJECTION
	);

	static TreeNodeIterator ni;
	static Vec transparentXf;
	static Vec opaqueXf;

	static bool init = true;
	if( init ) {
		init_TreeNodeIterator( &ni );
		init_Vec( &transparentXf, sizeof(Raw_Xformer*), ALIGN_DEFAULT );
		init_Vec( &opaqueXf, sizeof(Raw_Xformer*), ALIGN_DEFAULT );
		init = false;
	}
	reset_TreeNodeIterator( &ni, scene );

	// gather TreeNode pointers into a list
	static TreeNode* tn_list[1+MAX_NODES];
	{
		int i = 0;
		for( TreeNode* tn; NULL != (tn = inc_TreeNodeIterator( &ni )); ++i ) {
			tn_list[i] = tn;
		}
		tn_list[i] = NULL;
	}

	// 0th pass - split objects into transparent and non-transparent
	reset_Vec( &opaqueXf );
	reset_Vec( &transparentXf );
	for( TreeNode** it = tn_list; NULL != *it; ++it ) {
		Raw_Xformer* raw_xf = getData_TreeNode( *it );
		switch( raw_xf->tag ) {
			case OBJECT_BILLBOARD: {
				if( mesh_is_opaque( &( (Billboard_Xformer*)raw_xf )->billboard->mesh ) ) {
					pushBack_Vec( &opaqueXf, &raw_xf );
				} else {
					pushBack_Vec( &transparentXf, &raw_xf );
				}
				break;
			}
			case OBJECT_MESH: {
				if( mesh_is_opaque( ( (Mesh_Xformer*)raw_xf )->mesh ) ) {
					pushBack_Vec( &opaqueXf, &raw_xf );
				} else {
					pushBack_Vec( &transparentXf, &raw_xf );
				}
				break;
			}
			default:
				break;
		}
	}

	Matrix* cpuMtx = getElement_Vec( &engine->cpuMatrixes, 0 );
	// 1st pass - calculate world coordintes of nodes
	for( TreeNode** it = tn_list; NULL != *it; ++it ) {
		Xformer* parent_xf = NULL;
		if( getParent_TreeNode( *it ) ) {
			parent_xf = ( (Raw_Xformer*)getData_TreeNode( getParent_TreeNode( *it ) ) )->xf;
		}
		finish_Xformer(
			( (Raw_Xformer*)getData_TreeNode( *it ) )->xf,
			parent_xf,
			cpuMtx
		);
		++cpuMtx;
	}

	// 2nd pass - calculate lighting
	for( TreeNode** it = tn_list; NULL != *it; ++it ) {
		const Raw_Xformer* raw_xf = getData_TreeNode( *it );

		switch( raw_xf->tag ) {
			case OBJECT_LUX: {
				pushLoadLight_DList(
					&engine->dl,
					( (Lux_Xformer*)raw_xf )->lux,
					&getCoordsys_Xformer( raw_xf->xf )->origin,
					//&getCoordsys_Xformer( raw_xf->xf )->n,
					0
				);
				break;
			}

			default:
				break;
		}
	}

	// 3rd pass - apply additional transformations
	for( TreeNode** it = tn_list; NULL != *it; ++it ) {
		Raw_Xformer* raw_xf = getData_TreeNode( *it );
		void* data = NULL;
		switch( raw_xf->tag ) {
			case OBJECT_BILLBOARD: {
				data = cam_xf->xf;
				break;
			}
			default:
				break;
		}

		if( raw_xf->xf->processor ) {
			raw_xf->xf->processor( (struct Xformer*)raw_xf->xf, data );
		}

		reset_Xformer( raw_xf->xf );
	}

	// 4th pass - build display list
	static Vec* const vec[] = { &opaqueXf, &transparentXf };
	for( uint_t vi=0; vi<sizeof(vec)/sizeof(*vec); ++vi ) {
		for( uint32_t i=0; i<getLength_Vec( vec[vi] ); ++i ) {
			Raw_Xformer* raw_xf = *(Raw_Xformer**)getElement_Vec( vec[vi], i );

			resetState( &engine -> dl );
			switch( raw_xf->tag ) {
				case OBJECT_BILLBOARD:
					handleBillboard( &engine->dl, (Billboard_Xformer*)raw_xf, cam_xf );
					break;
				case OBJECT_MESH: {
					handleMesh( &engine->dl, (Mesh_Xformer*)raw_xf, cam_xf );
					break;
				}

				default:
					break;
			}
		}
	}
}

void clearColorbufF_Engine( Engine* engine, void* arg ) {
	const Vector3f* color = arg;
	pushSetClearColor_DList( &engine->dl,
		color->x,
		color->y,
		color->z,
		1.0f
	);
	pushClearBuffers_DList( &engine->dl, ColorBuffer );
}

void clearZbufF_Engine( Engine* engine, void* arg ) {
	const float* zval = arg;
	pushSetClearZValue_DList( &engine->dl, *zval );
	pushClearBuffers_DList( &engine->dl, ZBuffer );
}

void processTask_Engine( Engine* engine, const Task* task ) {
	const Surface* cbuf = getRenderTarget_Engine( engine );
	const Surface* zbuf = getZBuffer_Engine( engine );

	reset_DList( &engine->dl );
	pushStart_DList( &engine->dl );
	pushSetViewport_DList( &engine->dl, getViewport_Engine( engine ) );
	pushSetBuffer_DList(
		&engine->dl,
		ColorBuffer,
		0,
		getWidth_Surface( cbuf ),
		getHeight_Surface( cbuf ),
		getData_Surface( cbuf )
	);
	pushSetBuffer_DList(
		&engine->dl,
		ZBuffer,
		0,
		getWidth_Surface( zbuf ),
		getHeight_Surface( zbuf ),
		getData_Surface( getZBuffer_Engine( engine ) )
	);

	task->f( engine, task->arg );

	pushSync_DList( &engine->dl );
	pushEnd_DList( &engine->dl );
}

Xformer* createXformer_Engine( Engine* self, ProcessFunc_Xformer processor ) {
	Xformer xf;
	init_Xformer( &xf, processor );
	pushBack_Vec( &self->xformers, &xf );

	return getElement_Vec(
		&self->xformers,
		getLength_Vec( &self->xformers ) - 1
	);
}

Raw_Xformer createRaw_Engine( Engine* self ) {
	Raw_Xformer raw_xf;

	raw_xf.tag = OBJECT_RAW;
	raw_xf.xf = createXformer_Engine( self, NULL );

	return raw_xf;
}

Mesh_Xformer createMesh_Engine( Engine* self, StaticMesh const* static_mesh ) {
	Mesh_Xformer mesh_xf;

	Mesh mesh;
	init_Mesh( &mesh, static_mesh );
	pushBack_Vec( &self->meshes, &mesh );
	mesh_xf.tag = OBJECT_MESH;
	mesh_xf.mesh = getElement_Vec(
		&self->meshes,
		getLength_Vec( &self->meshes ) - 1
	);

	mesh_xf.xf = createXformer_Engine( self, NULL );

	return mesh_xf;
}

Billboard_Xformer createBillboard_Engine( Engine* self, StaticMesh const* static_mesh ) {
	Billboard_Xformer bb_xf;

	Billboard bb;
	init_Billboard( &bb, static_mesh );
	pushBack_Vec( &self->billboards, &bb );

	bb_xf.tag = OBJECT_BILLBOARD;
	bb_xf.billboard = getElement_Vec(
		&self->billboards,
		getLength_Vec( &self->billboards ) - 1
	);

	bb_xf.xf = createXformer_Engine( self, processBillboard_Xformer );

	return bb_xf;
}

Camera_Xformer createCamera_Engine( Engine* self, CameraType type ) {
	Camera_Xformer cam_xf;

	Camera cam;
	init_Camera( &cam, type );
	pushBack_Vec( &self->cameras, &cam );

	cam_xf.tag = OBJECT_CAMERA;
	cam_xf.camera = getElement_Vec(
		&self->cameras,
		getLength_Vec( &self->cameras ) - 1
	);

	cam_xf.xf = createXformer_Engine( self, processCamera_Xformer );

	return cam_xf;
}

Lux_Xformer createLux_Engine( Engine* self ) {
	Lux_Xformer lux_xf;

	Lux lux;
	init_Lux( &lux );
	pushBack_Vec( &self->luxes, &lux );
	lux_xf.lux = getElement_Vec(
		&self->luxes,
		getLength_Vec( &self->luxes ) - 1
	);

	lux_xf.tag = OBJECT_LUX;
	lux_xf.xf = createXformer_Engine( self, NULL );

	return lux_xf;
}

